import os
from enum import Enum
from typing import List

import numpy as np

from funlbm.utils import deep_get


class BoundaryCondition(Enum):
    PERIODICAL = 11000
    WALL = 1200
    WALL_WITH_SPEED = 1201
    FAR_FIELD = 1300
    NON_EQUILIBRIUM = 1400
    FULL_DEVELOPMENT = 1500

    @staticmethod
    def find(code):
        for value in BoundaryCondition.__iter__():
            if code == value.value or code == str(value.name):
                return value
        return BoundaryCondition.WALL


class BaseConfig(object):
    def __init__(self, *args, **kwargs):
        self.expand = {}
        self.expand.update(kwargs)

    def _from_json(self, config_json: dict, *args, **kwargs):
        pass

    def from_json(self, config_json: dict, *args, **kwargs):
        self.expand.update(kwargs)
        self._from_json(config_json, *args, **kwargs)
        return self

    def get(self, key, default=None):
        value = deep_get(self.expand, key)
        if value is None:
            return default


class Boundary(BaseConfig):
    def __init__(
        self, condition: BoundaryCondition = BoundaryCondition.WALL, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.condition: BoundaryCondition = condition
        self.poiseuille = None
        self.config = {}
        self.config.update(kwargs)

    def get(self, key):
        return self.config[key]

    def is_condition(self, condition: BoundaryCondition):
        return self.condition == condition

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.condition = deep_get(config_json, "code") or self.condition
        self.poiseuille = deep_get(config_json, "poiseuille")
        self.config.update(kwargs)


class BoundaryConfig(BaseConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.left = Boundary(BoundaryCondition.WALL)
        self.right = Boundary(BoundaryCondition.WALL)
        self.back = Boundary(BoundaryCondition.WALL)
        self.forward = Boundary(BoundaryCondition.WALL)
        self.bottom = Boundary(BoundaryCondition.WALL)
        self.top = Boundary(BoundaryCondition.WALL)

    def _from_json(self, config_json: dict, *args, **kwargs):
        if "left" in config_json.keys():
            self.left = Boundary().from_json(config_json["left"])
        if "right" in config_json.keys():
            self.right = Boundary().from_json(config_json["right"])
        if "back" in config_json.keys():
            self.left = Boundary().from_json(config_json["back"])
        if "forward" in config_json.keys():
            self.left = Boundary().from_json(config_json["forward"])

        if "bottom" in config_json.keys():
            self.left = Boundary().from_json(config_json["bottom"])
        if "top" in config_json.keys():
            self.left = Boundary().from_json(config_json["left"])


class FileConfig(BaseConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cache_dir = "./data"

    @property
    def vtk_path(self):
        path = f"{self.cache_dir}/vtk"
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        return path

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.cache_dir = deep_get(config_json, "cache_dir") or self.cache_dir


class CoordConfig(BaseConfig):
    def __init__(self, alpha=np.pi / 2, beta=0, gamma=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.center = [0, 0, 0]
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.center = deep_get(config_json, "center") or self.center
        self.alpha = deep_get(config_json, "alpha") or self.alpha
        self.beta = deep_get(config_json, "beta") or self.beta
        self.gamma = deep_get(config_json, "gamma") or self.gamma


class ParticleConfig(BaseConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.coord_config: CoordConfig = CoordConfig()

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.coord_config.from_json(deep_get(config_json, "coord"))


class FlowConfig(BaseConfig):
    def __init__(self, *args, **kwargs):
        self.size = np.zeros(3)
        self.boundary: BoundaryConfig = None
        super().__init__(*args, **kwargs)

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.size = np.array(
            deep_get(config_json, "size") or [100, 100, 100], dtype=int
        )
        self.boundary = BoundaryConfig().from_json(
            deep_get(config_json, "boundary") or {}
        )


class Config(BaseConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dt = 0.1
        self.dx = 0.1
        self.file_config = FileConfig()
        self.flow_config = FlowConfig()
        self.particles: List[ParticleConfig] = []

    def _from_json(self, config_json: dict, *args, **kwargs):
        self.dt = deep_get(config_json, "dt") or self.dt
        self.dx = deep_get(config_json, "dx") or self.dx
        self.file_config = FileConfig().from_json(deep_get(config_json, "file") or {})
        self.flow_config = FlowConfig().from_json(deep_get(config_json, "flow") or {})
        for config in deep_get(config_json, "particles") or []:
            self.particles.append(ParticleConfig().from_json(config_json=config))



import numpy as np
from funvtk.hl import gridToVTK, pointsToVTK
from tqdm import tqdm

from funlbm.config import Config
from funlbm.flow.flow3d import Flow3D
from funlbm.particle import Ellipsoid


class Solver(object):
    def __init__(self, config: Config, *args, **kwargs):
        self.config = config
        self.flow = Flow3D(config=config.flow_config)
        self.particles = [Ellipsoid(config=con) for con in config.particles]

    def run(self):
        self.init()
        total = int(9000 / self.config.dt)
        pbar = tqdm(range(total))
        pbar = range(total)
        for step in pbar:
            self.step(step)
            # print(f"{step}"
            #       f"\t{np.max(self.flow.f):8f}"
            #       f"\t{np.mean(self.flow.u[1, 1:-1, 1:-1, :]):8f}"
            #       f"\t{np.mean(self.flow.u[10, 1:-1, 1:-1, :]):8f}"
            #       f"\t{np.sum(self.flow.f):8f}"
            #       f"\t{self.particles[0].cx}"
            #       )

    def init(self):
        # 初始化流程
        # 初始化边界条件
        self.flow.init()
        # 初始化颗粒

        # self = self.particles[0]
        self.flow.cul_equ(tau=1)

        for particle in self.particles:
            particle.init()

    def flow_to_lagrange(self, n=2, h=1):
        for particle in self.particles:
            rl = np.array(np.floor(particle.lx) - (n - 1) * h, dtype=int)
            rl[rl < 0] = 0
            rr = rl + (2 * n - 1) * h + 1

            lu = np.zeros(particle.lu.shape)
            for index, lar in enumerate(particle.lx):
                i0, i1 = rl[index], rr[index]
                tmp = self.flow.x[i0[0] : i1[0], i0[1] : i1[1], i0[2] : i1[2], :] - lar
                tmp = (1 + np.cos(np.abs(tmp / h * np.pi / 2 / h))) / 4 / h
                tmp = np.prod(tmp, axis=-1, keepdims=True)

                lu[index, :] = np.sum(
                    self.flow.u[i0[0] : i1[0], i0[1] : i1[1], i0[2] : i1[2], :] * tmp
                )
                particle.lrou[index, :] = np.sum(
                    self.flow.rou[i0[0] : i1[0], i0[1] : i1[1], i0[2] : i1[2], :] * tmp
                )
                # print(lar, i1)
            u_theta = 0

            particle.lu[:, :] = (
                particle.cu + np.cross(particle.cw, particle.lx - particle.cx) + u_theta
            )
            particle.lF = particle.lrou * (particle.lu - lu)

            # TODO 力矩的公式到底是r×F，F×r
            # particle.lT = np.cross(particle.lF, particle.lx - particle.cx)
            particle.lT = np.cross(particle.lx - particle.cx, particle.lF)

    def lagrange_to_flow(self, n=2, h=1):
        for particle in self.particles:
            rl = np.array(np.floor(particle.lx) - (n - 1) * h, dtype=int)
            rl[rl < 0] = 0
            rr = rl + (2 * n - 1) * h + 1
            for index, lar in enumerate(particle.lx):
                i0, i1 = rl[index], rr[index]
                tmp = self.flow.x[i0[0] : i1[0], i0[1] : i1[1], i0[2] : i1[2], :] - lar
                tmp = (1 + np.cos(np.abs(tmp / h * np.pi / 2 / h))) / 4 / h
                tmp = np.prod(tmp, axis=-1, keepdims=True)
                tmp = tmp * particle.lF[index, :] * particle.lm[index]
                self.flow.FOL[i0[0] : i1[0], i0[1] : i1[1], i0[2] : i1[2], :] = tmp

    def particle_to_wall(self):
        k0 = 300
        for particle in self.particles:
            n = np.array(
                [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0], [0, -1, 0], [0, 0, -1]]
            )
            xt = np.concatenate(
                [
                    np.zeros(self.config.flow_config.size.shape),
                    self.config.flow_config.size,
                ]
            )
            xi = np.concatenate(
                [np.argmin(particle.lx, axis=0), np.argmax(particle.lx, axis=0)]
            )

            d = k0 * (
                1
                - abs(particle.lx[xi][[0, 1, 2, 3, 4, 5], [0, 1, 2, 0, 1, 2]] - xt)
                / (2 * self.config.dx)
            )
            d[d < 0] = 0
            if d.max() == 0:
                continue
            print(xi, d)
            particle.lF[xi] = particle.lF[xi] + np.multiply(n, np.expand_dims(d, 1))

    def step(self, step):
        # 流场碰撞
        self.flow.cul_equ()
        # 流场迁移
        self.flow.f_stream()
        # 流场计算-速度&密度
        self.flow.update_u_rou()

        # 浸没计算-流场->拉格朗日点
        self.flow_to_lagrange()

        self.particle_to_wall()

        # 浸没计算-拉格朗日点->颗粒
        [particle.update_from_lar(dt=self.config.dt) for particle in self.particles]

        # 浸没计算-拉格朗日点->流场
        self.lagrange_to_flow()

        # 流场二次碰撞
        self.flow.cul_equ2()
        # 流场计算-速度&密度
        self.flow.update_u_rou()

        # 颗粒->拉格朗日点
        [particle.update() for particle in self.particles]
        [print(particle.to_str()) for particle in self.particles]
        if step % 10 == 0 and step > 10:
            self.save(step)

    def save(self, step=10):

        shape = self.flow.u.shape
        xf, yf, zf = np.meshgrid(
            range(shape[0]),
            range(shape[1]),
            range(shape[2]),
            indexing="ij",
            sparse=False,
        )

        point_data = {
            "u": (
                self.flow.u[:, :, :, 0],
                self.flow.u[:, :, :, 1],
                self.flow.u[:, :, :, 2],
            ),
            # "rou": (self.rou[:, :, :, 0], self.rou[:, :, :, 0], self.rou[:, :, :, 0]),
        }

        gridToVTK(
            f"{self.config.file_config.vtk_path}/flow_" + str(step).zfill(10),
            xf,
            yf,
            zf,
            pointData=point_data,
        )

        for i, particle in enumerate(self.particles):
            xf, yf, zf = particle.lx[:, 0], particle.lx[:, 1], particle.lx[:, 2]
            data = {
                # "u": self.lu
                "u": particle.lu[:, 0],  # , self.lu[:, 1], self.lu[:, 2])
                # "u": (self.lu[:, 0], self.lu[:, 1], self.lu[:, 2])
            }
            fname = f"{self.config.file_config.vtk_path}/particle_{str(i).zfill(3)}_{str(step).zfill(10)}"
            pointsToVTK(fname, xf, yf, zf, data=data)



import numpy as np

from funlbm.config import FlowConfig
from funlbm.parameter import Par3D


class Flow(object):
    def __init__(self, config=None, *args, **kwargs):
        # 坐标
        self.x: np.ndarray = None
        # 力密度
        self.f: np.ndarray = None
        # 速度
        self.u: np.ndarray = None
        # 压强
        self.p: np.ndarray = None
        # 密度
        self.rou: np.ndarray = None
        # 剪切率相关的变量
        self.gama: np.ndarray = None

        self.FOL: np.ndarray = None

        self.tau: np.ndarray = None

        self.param = Par3D()
        self.config = config or FlowConfig()
